import os
import json

import singer
import singer.bookmarks as bookmarks
import singer.metrics as metrics
from singer import metadata
import tableauserverclient as TSC

from .tableau.datasources import get_all_datasource_details
from .tableau.groups import get_all_group_details
from .tableau.projects import get_all_project_details
from .tableau.schedules import get_all_schedule_details
from .tableau.tasks import get_all_task_details
from .tableau.workbooks import get_all_workbook_details


logger = singer.get_logger()

REQUIRED_CONFIG_KEYS = ['host']
KEY_PROPERTIES = {
    'connections': ['id'],
    'datasources': ['id'],
    'groups': ['id'],
    'projects': ['id'],
    'schedules': ['id'],
    'tasks': ['id'],
    'users': ['id'],
    'workbooks': ['id']
}


def get_bookmark(state, stream_name, bookmark_key, start_date):
    repo_stream_dict = bookmarks.get_bookmark(state, stream_name, bookmark_key)
    if repo_stream_dict:
        return repo_stream_dict.get(bookmark_key)
    if start_date:
        return start_date
    return None


def get_all_datasources(schema, server, authentication, state, mdata, start_date):
    bookmark_value = get_bookmark(state, 'datasources', 'updated_at', start_date)
    if bookmark_value:
        max_record_value = singer.utils.strptime_to_utc(bookmark_value)
    else:
        max_record_value = singer.utils.strptime_to_utc("1970-01-01")
    with metrics.record_counter('datasources') as counter:
        extraction_time = singer.utils.now()
        datasource_details = get_all_datasource_details(server=server, authentication=authentication, start_date=max_record_value)
        for datasource in datasource_details['datasources']:
            with singer.Transformer() as transformer:
                rec = transformer.transform(datasource, schema, metadata=metadata.to_map(mdata))
            singer.write_record('datasources', rec, time_extracted=extraction_time)
            counter.increment()
            if schema.get('connections'):
                for connection in datasource_details['connections']:
                    with singer.Transformer() as transformer:
                        rec = transformer.transform(connection, schema, metadata=metadata.to_map(mdata))
                    singer.write_record('connections', rec, time_extracted=extraction_time)
            if singer.utils.strptime_to_utc(rec['updated_at']) > max_record_value:
                max_record_value = singer.utils.strptime_to_utc(rec['updated_at'])
    state = singer.write_bookmark(state, 'datasources', 'updated_at', singer.utils.strftime(max_record_value))
    return state


def get_all_groups(schema, server, authentication, state, mdata, _start_date):
    with metrics.record_counter('groups') as counter:
        extraction_time = singer.utils.now()
        group_details = get_all_group_details(server=server, authentication=authentication)
        for group in group_details['groups']:
            with singer.Transformer() as transformer:
                rec = transformer.transform(group, schema, metadata=metadata.to_map(mdata))
            singer.write_record('groups', rec, time_extracted=extraction_time)
            counter.increment()
            if schema.get('users'):
                for user in group_details['users']:
                    with singer.Transformer() as transformer:
                        rec = transformer.transform(user, schema, metadata=metadata.to_map(mdata))
                    singer.write_record('users', rec, time_extracted=extraction_time)
    return state


def get_all_projects(schema, server, authentication, state, mdata, _start_date):
    with metrics.record_counter('projects') as counter:
        extraction_time = singer.utils.now()
        project_details = get_all_project_details(server=server, authentication=authentication)
        for project in project_details:
            with singer.Transformer() as transformer:
                rec = transformer.transform(project, schema, metadata=metadata.to_map(mdata))
            singer.write_record('projects', rec, time_extracted=extraction_time)
            counter.increment()
    return state


def get_all_schedules(schema, server, authentication, state, mdata, _start_date):
    with metrics.record_counter('schedules') as counter:
        extraction_time = singer.utils.now()
        schedule_details = get_all_schedule_details(server=server, authentication=authentication)
        for schedule in schedule_details:
            with singer.Transformer() as transformer:
                rec = transformer.transform(schedule, schema, metadata=metadata.to_map(mdata))
            singer.write_record('schedules', rec, time_extracted=extraction_time)
            counter.increment()
    return state


def get_all_tasks(schema, server, authentication, state, mdata, _start_date):
    with metrics.record_counter('tasks') as counter:
        extraction_time = singer.utils.now()
        task_details = get_all_task_details(server=server, authentication=authentication)
        for task in task_details:
            with singer.Transformer() as transformer:
                rec = transformer.transform(task, schema, metadata=metadata.to_map(mdata))
            singer.write_record('tasks', rec, time_extracted=extraction_time)
            counter.increment()
    return state


def get_all_workbooks(schema, server, authentication, state, mdata, start_date):
    bookmark_value = get_bookmark(state, 'workbooks', 'updated_at', start_date)
    if bookmark_value:
        max_record_value = singer.utils.strptime_to_utc(bookmark_value)
    else:
        max_record_value = singer.utils.strptime_to_utc("1970-01-01")
    with metrics.record_counter('workbooks') as counter:
        extraction_time = singer.utils.now()
        workbook_details = get_all_workbook_details(server=server, authentication=authentication, start_date=max_record_value)
        for workbook in workbook_details['workbooks']:
            with singer.Transformer() as transformer:
                rec = transformer.transform(workbook, schema, metadata=metadata.to_map(mdata))
            singer.write_record('workbooks', rec, time_extracted=extraction_time)
            counter.increment()
        if singer.utils.strptime_to_utc(rec['updated_at']) > max_record_value:
            max_record_value = singer.utils.strptime_to_utc(rec['updated_at'])
    state = singer.write_bookmark(state, 'workbooks', 'updated_at', singer.utils.strftime(max_record_value))
    return state


SYNC_FUNCTIONS = {
    'datasources': get_all_datasources,
    'groups': get_all_groups,
    'projects': get_all_projects,
    'schedules': get_all_schedules,
    'tasks': get_all_tasks,
    'workbooks': get_all_workbooks
}

SUB_STREAMS = {
    'datasources': ['connections'],
    'groups': ['users']  ## should users be a sub-stream of groups?
}


def get_abs_path(path):
    return os.path.join(os.path.dirname(os.path.realpath(__file__)), path)


def load_schemas():
    schemas = {}
    for filename in os.listdir(get_abs_path('schemas')):
        path = get_abs_path('schemas') + '/' + filename
        file_raw = filename.replace('.json', '')
        with open(path) as file:
            schemas[file_raw] = json.load(file)
    return schemas


def populate_metadata(schema_name, schema):
    mdata = metadata.new()
    mdata = metadata.write(mdata, (), 'table-key-properties', KEY_PROPERTIES[schema_name])
    for field_name in schema['properties'].keys():
        if field_name in KEY_PROPERTIES[schema_name]:
            mdata = metadata.write(mdata, ('properties', field_name), 'inclusion', 'automatic')
        else:
            mdata = metadata.write(mdata, ('properties', field_name), 'inclusion', 'available')
    return mdata


def get_catalog():
    raw_schemas = load_schemas()
    streams = []
    for schema_name, schema in raw_schemas.items():
        # get metadata for each field
        mdata = populate_metadata(schema_name, schema)
        # create and add catalog entry
        catalog_entry = {
            'stream': schema_name,
            'tap_stream_id': schema_name,
            'schema': schema,
            'metadata': metadata.to_list(mdata),
            'key_properties': KEY_PROPERTIES[schema_name],
        }
        streams.append(catalog_entry)
    return {'streams': streams}


def get_selected_streams(catalog):
    '''
    Gets selected streams.  Checks schema's 'selected'
    first -- and then checks metadata, looking for an empty
    breadcrumb and mdata with a 'selected' entry
    '''
    selected_streams = []
    for stream in catalog['streams']:
        stream_metadata = stream['metadata']
        if stream['schema'].get('selected', False):
            selected_streams.append(stream['tap_stream_id'])
        else:
            for entry in stream_metadata:
                # stream metadata will have empty breadcrumb
                if not entry['breadcrumb'] and entry['metadata'].get('selected', None):
                    selected_streams.append(stream['tap_stream_id'])
    return selected_streams


def get_stream_from_catalog(stream_id, catalog):
    for stream in catalog['streams']:
        if stream['tap_stream_id'] == stream_id:
            return stream
    return None


def do_discover(config):
    catalog = get_catalog()
    # dump catalog
    print(json.dumps(catalog, indent=2))


def do_sync(config, state, catalog):
    if config.get('username') and config.get('password'):
        print("Using username/ password based authentication")
        authentication = TSC.TableauAuth(config['username'], config['password'], site_id=config.get('site_id'))
    elif config.get('token_name') and config.get('token'):
        print("Using token based authentication")
        authentication = TSC.PersonalAccessTokenAuth(config['token_name'], config['token'], site_id=config.get('site_id'))
    else:
        raise ValueError("Must specify username/ password or token_name/ token for authentication")
    server = TSC.Server(config['host'], use_server_version=True)

    start_date = config['start_date'] if 'start_date' in config else None
    selected_stream_ids = get_selected_streams(catalog)

    for stream in catalog['streams']:
        stream_id = stream['tap_stream_id']
        stream_schema = stream['schema']
        mdata = stream['metadata']
        if not SYNC_FUNCTIONS.get(stream_id):
            continue
        if stream_id in selected_stream_ids:
            singer.write_schema(stream_id, stream_schema, stream['key_properties'])
            sync_func = SYNC_FUNCTIONS[stream_id]
            sub_stream_ids = SUB_STREAMS.get(stream_id, None)
            if not sub_stream_ids:
                state = sync_func(stream_schema, server, authentication, state, mdata, start_date)
            else:
                stream_schemas = {stream_id: stream_schema}
                for sub_stream_id in sub_stream_ids:
                    if sub_stream_id in selected_stream_ids:
                        sub_stream = get_stream_from_catalog(sub_stream_id, catalog)
                        stream_schemas[sub_stream_id] = sub_stream['schema']
                        singer.write_schema(sub_stream_id, sub_stream['schema'], sub_stream['key_properties'])
                state = sync_func(stream_schema, server, authentication, state, mdata, start_date)
            singer.write_state(state)


@singer.utils.handle_top_exception(logger)
def main():
    args = singer.utils.parse_args(REQUIRED_CONFIG_KEYS)

    if args.discover:
        do_discover(args.config)
    else:
        catalog = args.catalog.to_dict() if args.catalog else get_catalog()
        do_sync(args.config, args.state, catalog)


if __name__ == '__main__':
    main()



import singer
import tableauserverclient as TSC


def format_datetime(dt):
    if dt is not None:
        return singer.utils.strftime(dt)
    return None


def get_start_date_filter(start_date):
    req_option = TSC.RequestOptions()
    req_option.sort.add(
        TSC.Sort(
            TSC.RequestOptions.Field.UpdatedAt,
            TSC.RequestOptions.Direction.Asc
        )
    )
    req_option.filter.add(
        TSC.Filter(
            TSC.RequestOptions.Field.UpdatedAt,
            TSC.RequestOptions.Operator.GreaterThanOrEqual,
            format_datetime(start_date)
        )
    )
    return req_option



def get_permission_details(permission):
    return {
        'capabilities': permission.capabilities,
        'grantee_id': permission.grantee.id,
        'grantee_tag_name': permission.grantee.tag_name
    }



from .connections import get_connection_details
from .permissions import get_permission_details
from .utils import format_datetime
from .utils import get_start_date_filter


def get_datasource_details(datasource):
    return {
        'ask_data_enablement': datasource.ask_data_enablement,
        'certification_note': datasource.certification_note,
        'certified': datasource.certified,
        'content_url': datasource.content_url,
        'created_at': format_datetime(datasource.created_at),
        'datasource_type': datasource.datasource_type,
        'description': datasource.description,
        'encrypt_extracts': datasource.encrypt_extracts,
        'has_extracts': datasource.has_extracts,
        'id': datasource.id,
        'name': datasource.name,
        'owner_id': datasource.owner_id,
        'permissions': [get_permission_details(permission) for permission in datasource.permissions],
        'project_id': datasource.project_id,
        'project_name': datasource.project_name,
        'tags': [tag for tag in datasource.tags],
        'updated_at': format_datetime(datasource.updated_at),
        'use_remote_query_agent': datasource.use_remote_query_agent,
        'webpage_url': datasource.webpage_url
    }


def get_all_datasources(server, start_date):
    filter = get_start_date_filter(start_date=start_date)
    all_datasources, _ = server.datasources.get(filter)
    for datasource in all_datasources:
        server.datasources.populate_connections(datasource)
        server.datasources.populate_permissions(datasource)
    return all_datasources


def get_all_datasource_details(server, authentication, start_date):
    connections = []
    datasources = []
    if not server.is_signed_in():
        server.auth.sign_in(authentication)
    all_datasources = get_all_datasources(server=server, start_date=start_date)
    for datasource in all_datasources:
        datasources.append(get_datasource_details(datasource=datasource))
        for connection in datasource.connections:
            connections.append(get_connection_details(connection=connection))
    server.auth.sign_out()
    return {
        'datasources': datasources,
        'connections': connections,
    }



def get_user_details(user):
    return {
        'id': user.id,
        'auth_setting': user.auth_setting,
        'email': user.email,
        'name': user.name,
        'full_name': user.fullname,
        'role': user.site_role,
    }



from .connections import get_connection_details
from .permissions import get_permission_details
from .utils import format_datetime
from .utils import get_start_date_filter


def get_view_details(view):
    return {
        'content_url': view.content_url,
        'created_at': format_datetime(view.created_at),
        # 'csv': view.csv,
        'id': view.id,
        # 'image': view.image,
        'name': view.name,
        'owner_id': view.owner_id,
        # 'pdf': view.pdf,
        'permissions': [get_permission_details(permission) for permission in view.permissions],
        # 'preview_image': view.preview_image,
        'project_id': view.project_id,
        'sheet_type': view.sheet_type,
        'tags': view.tags,
        # 'total_views': view.total_views,
        'updated_at': format_datetime(view.updated_at),
        'workbook_id': view.workbook_id
    }


def get_workbook_details(workbook):
    return {
        # 'connections': [get_connection_details(connection) for connection in workbook.connections],
        'content_url': workbook.content_url,
        'created_at': format_datetime(workbook.created_at),
        'data_acceleration_config': workbook.data_acceleration_config,
        'description': workbook.description,
        'id': workbook.id,
        'name': workbook.name,
        'owner_id': workbook.owner_id,
        'permissions': [get_permission_details(permission) for permission in workbook.permissions],
    #     'preview_image': workbook.preview_image,
        'project_id': workbook.project_id,
        'project_name': workbook.project_name,
        'show_tabs': workbook.show_tabs,
        'size': workbook.size,
        'tags': [tag for tag in workbook.tags],
        'updated_at': format_datetime(workbook.updated_at),
        # 'views': workbook.views,
        'webpage_url': workbook.webpage_url
    }


def get_all_workbooks(server, start_date):
    filter = get_start_date_filter(start_date=start_date)
    all_workbooks, _ = server.workbooks.get(filter)
    for workbook in all_workbooks:
        server.workbooks.populate_connections(workbook)
        server.workbooks.populate_permissions(workbook)
        server.workbooks.populate_views(workbook)
    return all_workbooks


def get_all_workbook_details(server, authentication, start_date):
    workbooks = []
    connections = []
    if not server.is_signed_in():
        server.auth.sign_in(authentication)
    all_workbooks = get_all_workbooks(server=server, start_date=start_date)
    for workbook in all_workbooks:
        workbooks.append(get_workbook_details(workbook=workbook))
        for connection in workbook.connections:
            connections.append(get_connection_details(connection=connection))
    server.auth.sign_out()
    return {
        'workbooks': workbooks,
        'connections': connections,
    }



from .users import get_user_details


def get_group_details(group):
    return {
        'domain_name': group.domain_name,
        'id': group.id,
        'license_mode': group.license_mode,
        'minimum_site_role': group.minimum_site_role,
        'name': group.name,
        'tag_name': group.tag_name,
        'users': [user for user in group.users]
    }


def get_all_groups(server):
    all_groups, _ = server.groups.get()
    for group in all_groups:
        server.groups.populate_users(group)
    return all_groups


def get_all_group_details(server, authentication):
    groups = []
    users = []
    if not server.is_signed_in():
        server.auth.sign_in(authentication)
    all_groups = get_all_groups(server=server)
    for group in all_groups:
        group_details = get_group_details(group=group)
        groups.append(group_details)
        for user in group.users:
            if user.id not in set(user['id'] for user in users):
                users.append(get_user_details(user=user))
        group_details.pop('users')
    server.auth.sign_out()
    return {
        'groups': groups,
        'users': users,
    }



from .permissions import get_permission_details


def get_project_details(project, server, authentication):
    with server.auth.sign_in(authentication):
        return {
            'content_permissions': project.content_permissions,
            'default_datasource_permissions': [get_permission_details(permission) for permission in project.default_datasource_permissions],
            'default_flow_permissions': [get_permission_details(permission) for permission in project.default_flow_permissions],
            'default_workbook_permissions': [get_permission_details(permission) for permission in project.default_workbook_permissions],
            'description': project.description,
            'id': project.id,
            'is_default': project.is_default(),
            'name': project.name,
            'owner_id': project.owner_id,
            'parent_id': project.parent_id
        }


def get_all_projects(
        server,
        authentication,
        populate_permissions=True,
        populate_datasource_default_permissions=True,
        populate_flow_default_permissions=True,
        populate_workbook_default_permissions=True
):
    with server.auth.sign_in(authentication):
        all_projects, _ = server.projects.get()
        for project in all_projects:
            if populate_permissions:
                server.projects.populate_permissions(project)
            if populate_datasource_default_permissions:
                server.projects.populate_datasource_default_permissions(project)
            if populate_flow_default_permissions:
                server.projects.populate_flow_default_permissions(project)
            if populate_workbook_default_permissions:
                server.projects.populate_workbook_default_permissions(project)
    return all_projects


def get_all_project_details(server, authentication):
    projects = []
    with server.auth.sign_in(authentication):
        all_projects = get_all_projects(server=server, authentication=authentication)
        for project in all_projects:
            projects.append(get_project_details(project=project, server=server, authentication=authentication))
    return projects



from .utils import format_datetime


def get_target_details(target):
    return {
        'id': target.id,
        'type': target.type
    }


def get_task_details(task):
    return {
        'consecutive_failed_count': task.consecutive_failed_count,
        'id': task.id,
        'last_run_at': format_datetime(task.last_run_at),
        'priority': task.priority,
        'schedule_id': task.schedule_id,
        'target': get_target_details(target=task.target),
        'task_type': task.task_type
    }


def get_all_tasks(server):
    all_tasks, _ = server.tasks.get()
    return all_tasks


def get_all_task_details(server, authentication):
    if not server.is_signed_in():
        server.auth.sign_in(authentication)
    all_tasks = get_all_tasks(server=server)
    tasks = []
    for task in all_tasks:
        tasks.append(get_task_details(task=task))
    return tasks



def get_connection_details(connection):
    return {
        'connection_type': connection.connection_type,
        'connection_credentials': connection.connection_credentials,
        'datasource_id': connection.datasource_id,
        'datasource_name': connection.datasource_name,
        'embed_password': connection.embed_password,
        'id': connection.id,
        'server_address': connection.server_address,
        'server_port': connection.server_port,
        'username': connection.username
    }



from .utils import format_datetime


def get_schedule_details(schedule):
    return {
        'created_at': format_datetime(schedule.created_at),
        'end_schedule_at': format_datetime(schedule.end_schedule_at),
        'execution_order': schedule.execution_order,
        'id': schedule.id,
        'interval_item': schedule.interval_item,
        'name': schedule.name,
        'next_run_at': format_datetime(schedule.next_run_at),
        'priority': schedule.priority,
        'schedule_type': schedule.schedule_type,
        'state': schedule.state,
        'updated_at': format_datetime(schedule.updated_at),
    }


def get_all_schedules(server):
    all_schedules, _ = server.schedules.get()
    return all_schedules


def get_all_schedule_details(server, authentication):
    if not server.is_signed_in():
        server.auth.sign_in(authentication)
    all_schedules = get_all_schedules(server=server)
    schedules = []
    for schedule in all_schedules:
        schedules.append(get_schedule_details(schedule=schedule))
    return schedules



#!/usr/bin/env python
import pathlib
from setuptools import setup
# The directory containing this file

setup(
    name='tap-tableau',
    version='0.0.2',
    description='Singer tap for extracting data from the Tableau API',
    long_description=(pathlib.Path(__file__).parent / "README.md").read_text(),
    long_description_content_type="text/markdown",
    author='Beauty Pie',
    url='https://beautypie.com',
    classifiers=['Programming Language :: Python :: 3 :: Only'],
    install_requires=[
        'singer-python==5.12.1',
        'requests==2.25.0',
        'tableauserverclient==0.15.0'
    ],
    extras_require={
        'dev': [
            'pylint',
            'nose',
        ]
    },
    entry_points='''
        [console_scripts]
        tap-tableau=tap_tableau:main
    ''',
    packages=['tap_tableau'],
    package_data={
        'tap_tableau': ['tap_tableau/schemas/*.json']
    },
    include_package_data=True
)
